import 'package:copy_with_extension/copy_with_extension.dart';
import 'package:healthrian_ble_interface/healthrian_ble_interface.dart';
import 'package:healthrian_ble_interface/src/extention/extension.dart';
import 'package:lzma/lzma.dart';
import 'package:rxdart/rxdart.dart';

part 'module_ecg.g.dart';

@CopyWith()
class ModuleEcg {
  ModuleEcg({
    required this.ble,
    this.commands = const Commands(),
    this.udfRawStream = Builtin.udfRawStream,
    this.udfConvert = Builtin.udfConvertWearEcg,
    this.udfLeadStream = Builtin.udfLeadStreamWearEcg,
    this.udfLeadCompute = Builtin.udfLeadComputeWearEcg,
    this.udfFilter = Builtin.udfFilter,
    this.samplingRate = 250,
    this.channelCount = 12,
    this.sampleCount = 9,
    this.filterCount = 8,
  });

  final ModuleBle ble;
  final Commands commands;

  final Stream<List<int>> Function({
    required ModuleEcg ecg,
  }) udfRawStream;

  final List<List<double>>? Function({
    required List<int> event,
    required int channelCount,
    required int sampleCount,
  }) udfConvert;

  final Stream<List<int>> Function({
    required ModuleEcg ecg,
  }) udfLeadStream;

  final LeadState? Function({
    required List<int> event,
  }) udfLeadCompute;

  final Channels Function({
    required Channels channels,
    required int channelCount,
    required int sampleCount,
    required DataFilter dataFilter,
    required bool isAttached,
  }) udfFilter;

  /// [samplingRate] is the number samples per second
  final int samplingRate;

  /// [channelCount] is the number of channels
  final int channelCount;

  /// [sampleCount] is the number of samples in a packet
  final int sampleCount;

  /// [filterCount] is the number of data filters
  final int filterCount;

  bool isStrictCallback = false;
  bool isCleanerCallback = false;

  /// [maxBuffer] decides the maximum length of buffered Stream.
  double maxBuffer = 40;

  int get _maxBufferSize => (maxBuffer * samplingRate).round();

  /// [lengthBuffer] decides a length of buffered Stream,
  double lengthBuffer = 30;

  int get _lengthBufferSize => (lengthBuffer * samplingRate).round();

  /// [displayBuffer] decides a length of buffered Stream,
  /// which is shown to screen.
  double displayBuffer = 5;

  int get _displayBufferSize => (displayBuffer * samplingRate).round();

  /// [refreshBuffer] decides a length to be hidden after refreshed.
  double refreshBuffer = 0.5;

  int get _refreshBufferSize => (refreshBuffer * samplingRate).round();

  /// [isFiltered] decides whether ecg signal is filtered.
  bool isFiltered = true;

  /// If [isCompressed] is `true`,
  /// ecg signal is decompressed by LZMA algorithm.
  bool isCompressed = false;

  DataFilter dataFilter = DataFilter(8);

  DataFilter? dataFilterSnapshot;

  AdaptiveHeartRate adaptiveHR = AdaptiveHeartRate();

  final List<(void Function(Channels, DataFilter)? callback, int period)>
      callbacks = [];

  /// [lastChannel] is a latest stream value from [ble.streams.bufferedEcg]
  Channels? lastChannel;

  /// [lastLeadState] is a latest stream value from [ble.streams.leadStateEcg]
  LeadState? lastLeadState;

  bool _isPaused = false;
  bool _isMeasuring = false;

  bool get isPaused => _isPaused;

  bool get isMeasuring => _isMeasuring;

  ///////////////////////////////////////////////////////////////////////

  void initStream() {
    ble.streams
      ..rawEcg = null
      ..lzmaEcg = null
      ..convertedEcg = null
      ..sampleEcg = null
      ..eventEcg = null
      ..leadStateEcg = null
      ..heartRateEcg = null
      ..bufferedEcg = null
      ..bufferedFilterEcg = null
      ..plotEcg = null
      ..plotRTEcg = null;
  }

  Future<void> start() async {
    dataFilter = DataFilter(filterCount, is250Hz: samplingRate == 250);
    adaptiveHR = AdaptiveHeartRate(is250Hz: samplingRate == 250);

    await _sendDate();
    await _sendStartSave();
    await _sendStart();

    _isMeasuring = true;
    _isPaused = false;

    ble.streams
      ..rawEcg = _getRaw().broadcast
      ..lzmaEcg = _getLzma().broadcast
      ..eventEcg = _getEvent()?.broadcast
      ..convertedEcg = _getConverted().broadcast
      ..sampleEcg = _getSampled().broadcast
      ..leadStateEcg = _getLeadState().broadcast
      ..heartRateEcg = _getHeartRate().broadcast
      ..bufferedEcg = _getBuffered().broadcast
      ..bufferedFilterEcg = _getBufferedFilter().broadcast
      ..plotEcg = _getPlot().broadcast
      ..plotRTEcg = _getPlotRT().broadcast;
  }

  void resetPlot() {
    ble.streams
      ..plotEcg = _getPlot().broadcast
      ..plotRTEcg = _getPlotRT().broadcast;
  }

  void saveFilterSnapshot() {
    dataFilterSnapshot = dataFilter;
  }

  void loadFilterSnapshot() {
    if (dataFilterSnapshot == null) return;
    dataFilter = dataFilterSnapshot!;
  }

  /// [pause]
  Future<void> pause() async {
    _isPaused = !_isPaused;
  }

  /// [stop] sends stop_ecg_packet to device.
  Future<void> stop() async {
    _isMeasuring = false;
    final control = ble.points[Characteristics.STOP_ECG];
    await control.writePacket(commands.stopEcg);
  }

  Future<void> _sendStart() async {
    final control = ble.points[Characteristics.START_ECG];
    await control.writePacket(commands.startEcg);
  }

  Future<void> _sendDate() async {
    final control = ble.points[Characteristics.SET_DATA];
    await control.writePacket(commands.setDate);
  }

  Future<void> _sendStartSave() async {
    final control = ble.points[Characteristics.CONTROL_POINT];
    await control.writePacket(commands.save.startEcg);
  }

  Stream<List<int>> _getRaw() async* {
    final events = udfRawStream(ecg: this);

    await for (final res in events) {
      if (!_isMeasuring) break;
      if (!ble.isSimulator && ble.device == null) break;
      if (ble.isSimulator && !ble.isSimulatorConnected) break;
      if (_isPaused) continue;
      yield res;
    }
  }

  Stream<List<int>> _getLzma() async* {
    final events = ble.streams.rawEcg;
    if (events == null) return;
    await for (final event in events) {
      yield lzma.decode(event);
    }
  }

  Stream<List<int>>? _getEvent() async* {
    final events = isCompressed ? ble.streams.lzmaEcg : ble.streams.rawEcg;
    if (events == null) return;
    await for (final event in events) {
      yield event;
    }
  }

  /// list double
  Stream<List<List<double>>> _getConverted() async* {
    final events = ble.streams.eventEcg;
    if (events == null) return;
    await for (final event in events) {
      final value = udfConvert(
        event: event,
        channelCount: channelCount,
        sampleCount: sampleCount,
      );
      if (value == null) continue;
      yield value;
    }
  }

  /// list double
  Stream<List<double>> _getSampled() async* {
    final events = ble.streams.eventEcg;
    if (events == null) return;

    yield* events
        .map((event) => udfConvert(
              event: event,
              channelCount: channelCount,
              sampleCount: sampleCount,
            ))
        .flatMap(
      (value) {
        final now = DateTime.now();

        /// TODO: 휴리스틱 - 다음 패킷이 먼저 도착하는 경우 발생하는 경우 방지
        final nextPacketTime =
            now.add(Duration(milliseconds: 1000 ~/ samplingRate * sampleCount));
        final list = [
          for (var i = 0; i < value!.first.length; i++)
            (i, [for (var j = 0; j < value.length; j++) value[j][i]]),
        ];
        final processedTime = DateTime.now();
        final tic = nextPacketTime.difference(processedTime).inMilliseconds ~/
            sampleCount;
        return Stream.fromIterable(list).flatMap(
          (sample) => TimerStream(
            sample.$2,
            Duration(milliseconds: tic * sample.$1),
          ),
        );
      },
    );
  }

  ///
  Stream<LeadState?> _getLeadState() async* {
    final events = udfLeadStream(ecg: this);
    var now = DateTime.now();

    await for (final event in events) {
      final currentTime = DateTime.now();
      if (currentTime.difference(now) <= const Duration(milliseconds: 500)) {
        continue;
      }

      now = currentTime;

      final leadState = udfLeadCompute(event: event);
      if (leadState == null) continue;
      yield leadState;

      lastLeadState = leadState;
    }
    yield null;
  }

  ///
  Stream<int?> _getHeartRate() async* {
    final events = ble.streams.convertedEcg;
    if (events == null) return;

    var lastSentTime = DateTime.now();

    await for (final event in events) {
      final l2 = event[event.length >= 2 ? 1 : 0];
      for (var i = 0; i < l2.length; i++) {
        final isAttached = lastLeadState?.isAttached ?? true;

        final output = adaptiveHR.calculateHR(
          l2[i],
          isAttached,
        );

        final hr = output <= 0 ? null : output;
        if (hr == null &&
            DateTime.now().difference(lastSentTime) <
                const Duration(seconds: 2)) continue;

        yield hr;
        lastSentTime = DateTime.now();
      }
    }
    yield null;
  }

  Stream<Channels> _getBuffered() async* {
    final events = ble.streams.convertedEcg;
    if (events == null) return;

    final res = List.generate(channelCount, (index) => <double>[]);

    var x = 0.0;

    final callbacksTargets = <int>[];

    await for (final event in events) {
      for (var k = 0; k < channelCount; k++) {
        res[k].addAll(event[k]);
      }

      for (final channel in res) {
        if (channel.length > _maxBufferSize) {
          channel.removeRange(0, channel.length - _lengthBufferSize);
        }
      }

      final channels = Channels.fromList(res);
      yield channels;

      if (callbacksTargets.isEmpty) {
        callbacksTargets.addAll(
          List.generate(callbacks.length, (index) => callbacks[index].$2),
        );
      }

      for (var i = 0; i < callbacks.length; i++) {
        final callback = callbacks[i].$1;
        final period = callbacks[i].$2;
        if (callback != null && x >= callbacksTargets[i]) {
          if (isStrictCallback) {
            final start = channels.l2.length - period - (x.round() % period);
            final end = channels.l2.length - (x.round() % period);
            callback.call(channels.range(start, end), dataFilter);
          } else {
            callback.call(channels, dataFilter);
          }
          callbacksTargets[i] += period;
        }
      }

      lastChannel = channels;
      x += sampleCount;
    }

    if (!isCleanerCallback) return;
    for (var i = 0; i < callbacks.length; i++) {
      final callback = callbacks[i].$1;
      final period = callbacks[i].$2;
      if (callback != null && x < callbacksTargets[i]) {
        if (isStrictCallback) {
          final start = lastChannel!.l2.length - (x.round() % period);
          final end = lastChannel!.l2.length;
          callback.call(lastChannel!.range(start, end), dataFilter);
        } else {
          callback.call(lastChannel!, dataFilter);
        }
        callbacksTargets[i] += period;
      }
    }
  }

  Stream<Channels> _getBufferedFilter() async* {
    final events = ble.streams.bufferedEcg;
    if (events == null) return;

    final res = List.generate(channelCount, (index) => <double>[]);

    await for (final event in events) {
      final isAttached = lastLeadState?.isAttached ?? true;
      final startIndex = event.l2.length - sampleCount;
      final endIndex = event.l2.length;
      for (var k = 0; k < channelCount; k++) {
        res[k].addAll(event.list[k].getRange(startIndex, endIndex));
      }
      for (final channel in res) {
        if (channel.length > _maxBufferSize) {
          channel.removeRange(0, channel.length - _lengthBufferSize);
        }
      }
      final channels = Channels.fromList(res);

      yield udfFilter(
        channels: channels,
        channelCount: channelCount,
        sampleCount: sampleCount,
        dataFilter: dataFilter,
        isAttached: isAttached,
      );
    }
  }

  /// OOO OOO OOO OOO NNN
  /// XXX OOO OOO OOO OOO
  ///     OOO OOO OOO OOO
  Stream<PlotChannels> _getPlot() async* {
    final events =
        isFiltered ? ble.streams.bufferedFilterEcg : ble.streams.bufferedEcg;

    if (events == null) return;

    final resList = List.generate(channelCount, (index) => <FlSpot>[]);
    final res =
        PlotChannels.fromList(resList, 0, _refreshBufferSize, sampleCount);
    var x = 0.0;

    await for (final event in events) {
      for (var i = event.l2.length - sampleCount; i < event.l2.length; i++) {
        final j = i - event.l2.length + sampleCount;

        for (var k = 0; k < event.list.length; k++) {
          res.list[k].add(FlSpot(x + j, event.list[k][i]));
        }
      }

      if (res.l2.length > event.l2.length) {
        for (var k = 0; k < event.list.length; k++) {
          res.list[k].removeRange(0, res.list[k].length - event.list[k].length);
        }
      }

      yield res;
      x += sampleCount;
    }
  }

  ///     NNN
  /// OOO XXX OOO OOO
  /// OOO OOO OOO OOO
  Stream<PlotChannels?> _getPlotRT() async* {
    final events = ble.streams.plotEcg;

    if (events == null) return;

    double skip(double x) => x % _displayBufferSize;
    FlSpot f(FlSpot original) => FlSpot(skip(original.x), original.y);

    final resList = List.generate(
      channelCount,
      (_) => List<FlSpot>.generate(
        _displayBufferSize,
        (index) => FlSpot.nullSpot,
        growable: false,
      ),
    );

    final res = PlotChannels.fromList(
      resList,
      0,
      _refreshBufferSize,
      sampleCount,
    );
    var x = 0;

    await for (final event in events) {
      final start = event.l2.length - sampleCount;
      final end = event.l2.length;
      for (var i = start; i < end; i++) {
        final j = i - event.l2.length + sampleCount;
        final index = (x + j.toDouble()).round() % _displayBufferSize;

        for (var k = 0; k < event.list.length; k++) {
          res.list[k][index] = f(event.list[k][i]);
        }
      }
      x += sampleCount;
      yield res.copyWith(pivot: (x % _displayBufferSize));
    }
    yield null;
  }
}
