import 'package:healthrian_ble_interface/healthrian_ble_interface.dart';
import 'package:healthrian_ble_interface/src/extention/extension.dart';

class Builtin {
  static bool scanFilterWearEcg(BluetoothDevice device) =>
      device.platformName.startsWith(DEVICE_NAME_WEAR_ECG_12);

  static bool scanFilterCardioClick(BluetoothDevice device) =>
      device.platformName.startsWith(DEVICE_NAME_CARDIO_CLICK);

  static bool scanFilterWearCardioT(BluetoothDevice device) =>
      device.platformName.startsWith(DEVICE_NAME_WEAR_CARDIO_T);

  static Stream<List<int>> udfRawStream({
    required ModuleEcg ecg,
  }) async* {
    final measurement = ecg.ble.points[Characteristics.MEASUREMENT_ECG];
    if (measurement == null) return;
    try {
      await measurement.setNotifyValue(true);
    } catch (e) {
      '_getStream : $e'.log;
    }
    await for (final event in measurement.lastValueStream) {
      if (event.isEmpty) continue;
      yield event;
    }
  }

  static List<List<double>>? udfConvertWearEcg({
    required List<int> event,
    required int channelCount,
    required int sampleCount,
  }) {
    final res = List.generate(
      channelCount,
      (_) => List.filled(sampleCount, 0.0),
      growable: false,
    );
    if (event.length < 4) return null;
    final id = event[3];
    if (id != 0x11) return null;

    /// [V6, L1, L2, V2, V3, V4, V5, V6, V1]
    final ordering = [11, 0, 1, 7, 8, 9, 10, 6];
    final skip = 6;
    for (var i = 0; i < sampleCount; i++) {
      for (var j = 0; j < 8; j++) {
        final msb = event[skip + 24 * i + j * 3];
        final mid = event[skip + 24 * i + j * 3 + 1];
        final lsb = event[skip + 24 * i + j * 3 + 2];
        final value = toDouble(msb, mid, lsb);
        res[ordering[j]][i] = value;
      }
    }

    /// l3, aVR, aVL, aVF
    for (var i = 0; i < sampleCount; i++) {
      final l1 = res[0][i];
      final l2 = res[1][i];
      res[2][i] = (l3(l1, l2));
      res[3][i] = (aVR(l1, l2));
      res[4][i] = (aVL(l1, l2));
      res[5][i] = (aVF(l1, l2));
    }
    return res;
  }

  static List<List<double>>? udfConvertCardioClick({
    required List<int> event,
    required int channelCount,
    required int sampleCount,
  }) {
    final res = List.generate(
      channelCount,
      (_) => List.filled(sampleCount, 0.0),
      growable: false,
    );
    if (event.isEmpty) return null;

    /// [V6, L1, L2, V2, V3, V4, V5, V6, V1]
    final ordering = [11, 0, 1, 7, 8, 9, 10, 6];
    final skip = 0;
    for (var i = 0; i < sampleCount; i++) {
      for (var j = 0; j < 8; j++) {
        final msb = event[skip + 16 * i + j * 2];
        final lsb = event[skip + 16 * i + j * 2 + 1];
        final value = toDoubleTwoByte(msb, lsb);
        res[ordering[j]][i] = value;
      }
    }
    for (var i = 0; i < sampleCount; i++) {
      final l1 = res[0][i];
      final l2 = res[1][i];
      res[2][i] = (l3(l1, l2));
      res[3][i] = (aVR(l1, l2));
      res[4][i] = (aVL(l1, l2));
      res[5][i] = (aVF(l1, l2));
    }
    return res;
  }

  static List<List<double>>? udfConvertWearCardioT({
    required List<int> event,
    required int channelCount,
    required int sampleCount,
  }) {
    final res = List.generate(
      channelCount,
      (_) => List.filled(sampleCount, 0.0),
      growable: false,
    );
    if (event.isEmpty) return null;

    for (var i = 0; i < sampleCount; i++) {
      for (var j = 0; j < channelCount; j++) {
        final msb = event[channelCount * 3 * i + j * 3];
        final mid = event[channelCount * 3 * i + j * 3 + 1];
        final lsb = event[channelCount * 3 * i + j * 3 + 2];
        final value = toDouble(msb, mid, lsb);
        res[j][i] = value;
      }
    }
    return res;
  }

  static Stream<List<int>> udfLeadStreamWearEcg({
    required ModuleEcg ecg,
  }) {
    return ecg.ble.streams.eventEcg!;
  }

  static Stream<List<int>> udfLeadStreamCardioClick({
    required ModuleEcg ecg,
  }) async* {
    final leadState = ecg.ble.points[Characteristics.LEAD_STATE_ECG];
    if (leadState == null) return;

    try {
      await leadState.setNotifyValue(true);
    } catch (e) {
      '_getLeadStateStream: $e'.log;
      try {
        await Future<void>.delayed(const Duration(seconds: 2));
        await leadState.setNotifyValue(true);
      } catch (e) {
        '_getLeadStateStream (second try): $e'.log;
      }
    }
    await for (final event in leadState.lastValueStream) {
      yield event;
    }
  }

  static Channels udfFilter({
    required Channels channels,
    required int channelCount,
    required int sampleCount,
    required DataFilter dataFilter,
    required bool isAttached,
  }) {
    final startIndex = channels.l2.length - sampleCount;
    for (var i = startIndex; i < channels.l2.length; i++) {
      final l1 = dataFilter.applyFilter(0, channels.l1[i], isAttached);
      final l2 = dataFilter.applyFilter(1, channels.l2[i], isAttached);

      channels.l1[i] = l1;
      channels.l2[i] = l2;
      channels.l3[i] = l3(l1, l2);

      channels.aVR[i] = aVR(l1, l2);
      channels.aVL[i] = aVL(l1, l2);
      channels.aVF[i] = aVF(l1, l2);

      channels.v1[i] = dataFilter.applyFilter(2, channels.v1[i], isAttached);
      channels.v2[i] = dataFilter.applyFilter(3, channels.v2[i], isAttached);
      channels.v3[i] = dataFilter.applyFilter(4, channels.v3[i], isAttached);
      channels.v4[i] = dataFilter.applyFilter(5, channels.v4[i], isAttached);
      channels.v5[i] = dataFilter.applyFilter(6, channels.v5[i], isAttached);
      channels.v6[i] = dataFilter.applyFilter(7, channels.v6[i], isAttached);
    }
    return channels;
  }

  static Channels udfFilterWearCardioT({
    required Channels channels,
    required int channelCount,
    required int sampleCount,
    required DataFilter dataFilter,
    required bool isAttached,
  }) {
    final startIndex = channels.l2.length - sampleCount;
    for (var i = startIndex; i < channels.l2.length; i++) {
      final l2 = dataFilter.applyFilter(1, channels.l2[i], isAttached);
      channels.l2[i] = l2;
    }
    return channels;
  }

  static LeadState? udfLeadComputeWearEcg({
    required List<int> event,
  }) {
    if (event.length < 4) return null;
    final id = event[3];
    if (id != 0x11) return null;
    if (event.length < 6) return null;
    final state1 = event[4];
    final state2 = event[5];

    final leadState = LeadState.fromBytes(state1, state2);
    return leadState;
  }

  static LeadState? udfLeadComputeCardioClick({
    required List<int> event,
  }) {
    print('udfLeadComputeCardioClick');
    if (event.length < 2) return null;
    final state1 = event[0];
    final state2 = event[1];

    final leadState = LeadState.fromBytes(state1, state2);
    return leadState;
  }

  static LeadState? udfLeadComputeWearCardioT({
    required List<int> event,
  }) {
    if (event.length < 4) return null;
    final leadState = LeadState(
      ra: event[3] & 0x01 == 0x01,
      rl: true,
      la: (event[3] >> 1) & 0x01 == 0x01,
      ll: true,
      v1: true,
      v2: true,
      v3: true,
      v4: true,
      v5: true,
      v6: true,
    );
    return leadState;
  }
}
